/*  $Header: /home/cvsroot/dvipdfmx/src/mpost.c,v 1.25 2004/08/07 09:52:34 hirata Exp $

    This is dvipdfmx, an eXtended version of dvipdfm by Mark A. Wicks.

    Copyright (C) 2002 by Jin-Hwan Cho and Shunsaku Hirata,
    the dvipdfmx project team <dvipdfmx@project.ktug.or.kr>

    Copyright (C) 1998, 1999 by Mark A. Wicks <mwicks@kettering.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <ctype.h>
#include <string.h>
#include <math.h>

#include "system.h"
#include "mem.h"
#include "error.h"
#include "mfileio.h"

#include "tfm.h"
#include "dvi.h"

#include "pdfobj.h"
#include "pdfspecial.h"
#include "pdfparse.h"
#include "pdflimits.h"
#include "pdfdev.h"
#include "pdfdoc.h"

#include "pdfcolor.h"
#include "pdfdraw.h"

#include "fontmap.h"
#include "subfont.h"

#include "pdfximage.h"

#include "mpost.h"

/*
 * In PDF, current path is not a part of graphics state parameter.
 * Hence, current path is not saved by the "q" operator  and is not
 * recovered by the "Q" operator. This means that the following PS
 * code
 *
 *   <path construction> gsave <path painting> grestore ...
 *
 * can't be translated to PDF code
 *
 *   <path construction> q <path painting> Q ...
 *
 * . Only clipping path (which is graphics state parameter in PDF
 * too) is treated in the same way. So, we write clipping path
 * immediately and forget about it but remember current path.
 */
#define MP_OPTIMIZE_PATH 0

static int mp_parse_body (char **start, char *end, double x_user, double y_user);

static struct mp_fonts 
{
  char * tex_name;
  int    font_id;
  int    tfm_id;     /* Used for text width calculation */
  int    subfont_id;
  double pt_size;
} *mp_fonts = NULL;

static int num_mp_fonts = 0, max_mp_fonts = 0;

/* Compatibility */
#define MP_CMODE_MPOST    0
#define MP_CMODE_DVIPSK   1
#define MP_CMODE_PTEXVERT 2
static int mp_cmode = MP_CMODE_MPOST;

static void
need_more_mp_fonts (unsigned n)
{
  if (num_mp_fonts + n > max_mp_fonts) {
    max_mp_fonts += MAX_FONTS;
    mp_fonts = RENEW(mp_fonts, max_mp_fonts, struct mp_fonts);
  }
}

static int
mp_locate_font (const char *tfm_name, double pt_size)
{
  int i, map_id, subfont_id;

  for (i = 0; i < num_mp_fonts; i++) {
    if (!strcmp(tfm_name, mp_fonts[i].tex_name) &&
	mp_fonts[i].pt_size == pt_size)
      return i;
  }

  need_more_mp_fonts(1);
  num_mp_fonts++;

  map_id = get_map_record(tfm_name, &subfont_id);

  mp_fonts[i].tex_name = NEW(strlen(tfm_name)+1, char);
  strcpy(mp_fonts[i].tex_name, tfm_name);
  mp_fonts[i].subfont_id = subfont_id;
  mp_fonts[i].pt_size = pt_size;

  mp_fonts[i].tfm_id  = tfm_open(tfm_name, 0); /* Need not exist in MP mode */

  /* The following line is a bit of a kludge.
   * MetaPost inclusion was an afterthought. */
  if (map_id < 0) {
    mp_fonts[i].font_id = dev_locate_font(tfm_name, (spt_t)(pt_size*dev_unit_dviunit()), map_id);
  } else {
    mp_fonts[i].font_id = dev_locate_font(fontmap_tex_name(map_id),
					  (spt_t)(pt_size*dev_unit_dviunit()), map_id);
  }
  if (mp_fonts[i].font_id < 0) {
    WARN("MPOST: No physical font assigned for \"%s\".", tfm_name);
  }

  return i;
}

static void
release_fonts (void)
{
  int i;

  for (i = 0; i < num_mp_fonts; i++) {
    RELEASE (mp_fonts[i].tex_name);
  }
  if (mp_fonts) {
    RELEASE (mp_fonts);
    mp_fonts = NULL;
  }
  num_mp_fonts = 0;
  max_mp_fonts = 0;
}

static int
mp_is_fontname (const char *token)
{
  int i, map_id, subfont_id;

  for (i = 0; i < num_mp_fonts; i++) {
    if (!strcmp(token, mp_fonts[i].tex_name)) {
      return 1;
    }
  }

  map_id = get_map_record(token, &subfont_id);
  if (map_id >= 0) {
    return 1;
  }

  return 0;
}

static int
scan_bbox (char **start, char *end, pdf_rect *bbox)
{
  char *llx = NULL, *lly = NULL, *urx = NULL, *ury = NULL;
  int error = 1;

  /* Scan for bounding box record */
  while (**start == '%') {
    if (!strncmp(*start, "%%BoundingBox:", 14)) {
      *start += 14;
      skip_white(start, end);
      /* Expect 4 numbers or fail */
      if ((llx = parse_number(start, end)) &&
	  (lly = parse_number(start, end)) &&
	  (urx = parse_number(start, end)) &&
	  (ury = parse_number(start, end))) {
	/* Set the crop box to the true bounding box specified in the file */
	bbox->llx = atof(llx);
	bbox->lly = atof(lly);
	bbox->urx = atof(urx);
	bbox->ury = atof(ury);
	error = 0;
      } else {
	WARN("Missing expected four numbers in BoundingBox.");
	dump(*start, end);
	error = 1;
      }
    }
    skip_line(start, end);
  }

  if (llx) RELEASE(llx);
  if (lly) RELEASE(lly);
  if (urx) RELEASE(urx);
  if (ury) RELEASE(ury);

  if (error) /* Didn't find Bounding box */
    WARN("Failed to find an expected BoundingBox record.");

  return error;
}

static void
skip_prolog (char **start, char *end)
{
  int   found_prolog = 0;
  char *save;

  save = *start;
  while (*start < end) {
    if (!strncmp(*start, "%%EndProlog", 11)) {
      found_prolog = 1;
      skip_line(start, end);
      break;
    } else if (!strncmp(*start, "%%Page:", 7)) {
      break;
    }
    skip_line(start, end);
  }
  if (!found_prolog)
    *start = save;
}

#define PDF_DASH_SIZE_MAX 16
typedef struct
{
  pdf_coord   cp;

  pdf_tmatrix matrix;   /* cm,  - */

  int font;             /* Tf,  - */

  pdf_color color;

  struct {
    int    num_dash;
    double pattern[PDF_DASH_SIZE_MAX];
    double offset;
  } line_dash;          /* d,  D  */

  double   line_width;  /* w,  LW */

  int      line_cap;    /* J,  LC */
  int      line_join;   /* j,  LJ */
  double   miter_limit; /* M,  ML */

  double   flatness;    /* i,  FL */
  int      intent;      /* ri, RI */

  struct {
    double stroke;     /* -,  CA */
    double fill;       /* -,  ca */
  } alpha;

  /* internal */
  pdf_path *path;
  int flags;
} pdf_xgstate;

#define PS_GSAVE_MAX 1024
pdf_xgstate mp_gstates[PS_GSAVE_MAX];
static int num_saves = -1;

#define CURRENT_GS()     (&(mp_gstates[num_saves]))
#define CURRENT_FONT()   (mp_gstates[num_saves].font)
#define CURRENT_PATH()   (mp_gstates[num_saves].path)
#define CURRENT_POINT()  (&(mp_gstates[num_saves].cp))
#define CURRENT_MATRIX() (&(mp_gstates[num_saves].matrix))
#define CURRENT_COLOR()  (&(mp_gstates[num_saves].color))

static void
init_gstate (pdf_xgstate *gs)
{
  gs->cp.x = 0.0;
  gs->cp.y = 0.0;

  gs->matrix.a = 1.0;
  gs->matrix.b = 0.0;
  gs->matrix.c = 0.0;
  gs->matrix.d = 1.0;
  gs->matrix.e = 0.0;
  gs->matrix.f = 0.0;

  gs->color.space_id  = PDF_COLOR_DEV_GRAY;
  gs->color.values[0] = 0.0;

  gs->font = -1;
  gs->path = pdf_draw_new_path();

  gs->line_dash.num_dash = 0;
  gs->line_dash.offset   = 0;
  gs->line_cap    = 0;
  gs->line_join   = 0;
  gs->line_width  = 1.0;
  gs->miter_limit = 10.0;
}

static void
set_currentpoint (double x, double y)
{
  pdf_coord *p;

  p = CURRENT_POINT();
  p->x = x;
  p->y = y;
}

static void
__concat (pdf_tmatrix *M)
{
  pdf_tmatrix *ctm;
  pdf_tmatrix M_i;

  pdf_copymatrix(&M_i, M);
  pdf_invertmatrix(&M_i);

  ctm = CURRENT_MATRIX();

  pdf_mulmatrix(ctm, M);
  pdf_draw_transform_path (CURRENT_PATH(),  &M_i);
  pdf_draw_transform_coord(CURRENT_POINT(), &M_i);
}

#define IS_IDENTITY_MATRIX(M) ((M)->a == 1.0 && (M)->b == 0.0 &&\
    (M)->c == 0.0 && (M)->d == 1.0 && (M)->e == 0.0 && (M)->f == 0.0)


/* PostScript Operators */

#define ADD          	1
#define SUB		2
#define MUL		3
#define DIV		4
#define NEG    	        5
#define TRUNCATE	6

#define CLEAR		10
#define EXCH		11
#define POP		12

#define NEWPATH		31
#define CLOSEPATH    	32
#define MOVETO		33
#define RMOVETO         34
#define CURVETO   	35
#define RCURVETO        36
#define LINETO		37
#define RLINETO		38
#define ARC             39
#define ARCN            40

#define FILL		41
#define STROKE		42
#define SHOW		43

#define CLIP         	44
#define EOCLIP         	45

#define SHOWPAGE	49

#define GSAVE		50
#define GRESTORE	51

#define CONCAT       	52
#define SCALE		53
#define TRANSLATE	54
#define ROTATE          55

#define SETLINEWIDTH	60
#define SETDASH		61
#define SETLINECAP 	62
#define SETLINEJOIN	63
#define SETMITERLIMIT	64

#define SETGRAY		70
#define SETRGBCOLOR	71
#define SETCMYKCOLOR	72

#define CURRENTPOINT    80
#define IDTRANSFORM	81
#define DTRANSFORM	82

#define FINDFONT        201
#define SCALEFONT       202
#define SETFONT         203
#define CURRENTFONT     204

#define FSHOW		1001
#define STEXFIG         1002
#define ETEXFIG         1003

static struct operators 
{
  const char *token;
  int         opcode;
} ps_operators[] = {
  {"add",          ADD},
  {"mul",          MUL},
  {"div",          DIV},
  {"neg",          NEG},
  {"sub",          SUB},  
  {"truncate",     TRUNCATE},

  {"clear",        CLEAR},
  {"exch",         EXCH},
  {"pop",          POP},

  {"clip",         CLIP},
  {"eoclip",       EOCLIP},
  {"closepath",    CLOSEPATH},
  {"concat",       CONCAT},

  {"newpath",      NEWPATH},
  {"moveto",       MOVETO},
  {"rmoveto",      RMOVETO},
  {"lineto",       LINETO},
  {"rlineto",      RLINETO},
  {"curveto",      CURVETO},
  {"rcurveto",     RCURVETO},
  {"arc",          ARC},
  {"arcn",         ARCN},

  {"stroke",       STROKE},  
  {"fill",         FILL},
  {"show",         SHOW},
  {"showpage",     SHOWPAGE},

  {"gsave",        GSAVE},
  {"grestore",     GRESTORE},
  {"translate",    TRANSLATE},
  {"rotate",       ROTATE},
  {"scale",        SCALE},

  {"setlinecap",    SETLINECAP},
  {"setlinejoin",   SETLINEJOIN},
  {"setlinewidth",  SETLINEWIDTH},
  {"setmiterlimit", SETMITERLIMIT},
  {"setdash",       SETDASH},

  {"setgray",      SETGRAY},
  {"setrgbcolor",  SETRGBCOLOR},
  {"setcmykcolor", SETCMYKCOLOR},

  {"currentpoint", CURRENTPOINT}, /* This is here for rotate support
				     in graphics package-not MP support */
  {"dtransform",   DTRANSFORM},
  {"idtransform",  IDTRANSFORM},

  {"findfont",     FINDFONT},
  {"scalefont",    SCALEFONT},
  {"setfont",      SETFONT},
  {"currentfont",  CURRENTFONT}
};

static struct operators mps_operators[] = {
  {"fshow",       FSHOW},
  {"startTexFig", STEXFIG},
  {"endTexFig",   ETEXFIG}
};

#define NUM_PS_OPERATORS  (sizeof(ps_operators)/sizeof(ps_operators[0]))
#define NUM_MPS_OPERATORS (sizeof(mps_operators)/sizeof(mps_operators[0]))
static int
get_opcode (const char *token)
{
  int i;

  for (i = 0; i < NUM_PS_OPERATORS; i++) {
    if (!strcmp(token, ps_operators[i].token)) {
      return ps_operators[i].opcode;
    }
  }

  for (i = 0; i < NUM_MPS_OPERATORS; i++) {
    if (!strcmp(token, mps_operators[i].token)) {
      return mps_operators[i].opcode;
    }
  }

  return -1;
}

#define PS_STACK_SIZE 1024

static pdf_obj *stack[PS_STACK_SIZE];
static unsigned top_stack = 0;

static int
do_gsave (void)
{
  pdf_xgstate *gs0, *gs1;
  int i;

  if (num_saves >= PS_GSAVE_MAX) {
    WARN("Too many gsaves.");
    return 1;
  }

  num_saves++;
  gs0 = &(mp_gstates[num_saves-1]);
  gs1 = &(mp_gstates[num_saves  ]);

  gs1->cp.x = gs0->cp.x;
  gs1->cp.y = gs0->cp.y;

  pdf_copymatrix(&gs1->matrix, &gs0->matrix);

  gs1->font = gs0->font;
  gs1->path = pdf_draw_new_path();
  pdf_draw_copy_path(gs1->path, gs0->path);

  gs1->line_dash.num_dash = gs0->line_dash.num_dash;
  for (i = 0; i < gs0->line_dash.num_dash; i++) {
    gs1->line_dash.pattern[i] = gs0->line_dash.pattern[i];
  }
  gs1->line_dash.offset = gs0->line_dash.offset;
  gs1->line_cap    = gs0->line_cap;
  gs1->line_join   = gs0->line_join;
  gs1->line_width  = gs0->line_width;
  gs1->miter_limit = gs0->miter_limit;

  memcpy(&gs1->color, &gs0->color, sizeof(pdf_color));

  return 0;
}

static int
do_grestore (void)
{
  pdf_xgstate *gs;

  if (num_saves <= 0) {
    WARN("Too many grestores.");
    return 1;
  }
  gs = CURRENT_GS();
  pdf_draw_release_path(gs->path);
  gs->path = NULL;

  num_saves--;
  dev_reselect_font();

  return 0;
}

#define POP_STACK()     ((top_stack > 0) ? stack[--top_stack] : NULL)
#define PUSH_STACK(o,e) { \
  if (top_stack < PS_STACK_SIZE) { \
    stack[top_stack++] = (o); \
  } else { \
    WARN("PS stack overflow including MetaPost file or inline PS code"); \
    *(e) = 1; \
  } \
}

static int
do_exch (void)
{
  pdf_obj *tmp;

  if (top_stack < 2)
    return -1;

  tmp = stack[top_stack-1];
  stack[top_stack-1] = stack[top_stack-2];
  stack[top_stack-2] = tmp;

  return 0;
}

static int
do_clear (void)
{
  pdf_obj *tmp;

  while (top_stack > 0) {
    tmp = POP_STACK();
    if (tmp)
      pdf_release_obj(tmp);
  }

  return 0;
}

static int
pop_get_numbers (double *values, int count)
{
  pdf_obj *tmp;

  while (count-- > 0) {
    tmp = POP_STACK();
    if (!tmp) {
      WARN("mpost: Stack underflow.");
      break;
    } else if (!PDF_OBJ_NUMBERTYPE(tmp)) {
      WARN("mpost: Not a number!");
      pdf_release_obj(tmp);
      break;
    }
    values[count] = pdf_number_value(tmp);
    pdf_release_obj(tmp);
  }

  return (count + 1);
}

static int
cvr_array (pdf_obj *array, double *values, int count)
{
  if (!PDF_OBJ_ARRAYTYPE(array)) {
    WARN("mpost: Not an array!");
  } else {
    pdf_obj *tmp;

    while (count-- > 0) {
      tmp = pdf_get_array(array, count);
      if (!PDF_OBJ_NUMBERTYPE(tmp)) {
	WARN("mpost: Not a number!");
	break;
      }
      values[count] = pdf_number_value(tmp);
    }
  }
  if (array)
    pdf_release_obj(array);

  return (count + 1);
}

static int
is_fontdict (pdf_obj *dict)
{
  pdf_obj *tmp;

  if (!PDF_OBJ_DICTTYPE(dict))
    return 0;

  tmp = pdf_lookup_dict(dict, "Type");
  if (!tmp || !PDF_OBJ_NAMETYPE(tmp) ||
      strcmp(pdf_name_value(tmp), "Font")) {
    return 0;
  }

  tmp = pdf_lookup_dict(dict, "FontName");
  if (!tmp || !PDF_OBJ_NAMETYPE(tmp)) {
    return 0;
  }

  tmp = pdf_lookup_dict(dict, "FontScale");
  if (!tmp || !PDF_OBJ_NUMBERTYPE(tmp)) {
    return 0;
  }

  return 1;
}

static int
do_findfont (void)
{
  int error = 0;
  pdf_obj *font_dict, *font_name;

  font_name = POP_STACK();
  if (!font_name)
    return 1;
  else if (PDF_OBJ_STRINGTYPE(font_name) ||
	   PDF_OBJ_NAMETYPE(font_name)) {
    /* Do not check the existence...
     * The reason for this is that we cannot locate PK font without
     * font scale.
     */
    font_dict = pdf_new_dict();
    pdf_add_dict(font_dict,
		 pdf_new_name("Type"), pdf_new_name("Font"));
    if (PDF_OBJ_STRINGTYPE(font_name)) {
      pdf_add_dict(font_dict,
		   pdf_new_name("FontName"),
		   pdf_new_name(pdf_string_value(font_name)));
      pdf_release_obj(font_name);
    } else {
      pdf_add_dict(font_dict,
		   pdf_new_name("FontName"), font_name);
    }
    pdf_add_dict(font_dict,
		 pdf_new_name("FontScale"), pdf_new_number(1.0));

    if (top_stack < PS_STACK_SIZE) {
      stack[top_stack++] = font_dict;
    } else {
      WARN("PS stack overflow including MetaPost file or inline PS code");
      pdf_release_obj(font_dict);
      error = 1;
    }
  } else {
    error = 1;
  }

  return error;
}

static int
do_scalefont (void)
{
  int error = 0;
  pdf_obj *font_dict;
  pdf_obj *font_scale;
  double   scale;

  error = pop_get_numbers(&scale, 1);
  if (error)
    return error;

  font_dict = POP_STACK();
  if (!font_dict)
    error = 1;
  else if (is_fontdict(font_dict)) {
    font_scale  = pdf_lookup_dict(font_dict, "FontScale");
    pdf_set_number(font_scale, pdf_number_value(font_scale)*scale);
    if (top_stack < PS_STACK_SIZE) {
      stack[top_stack++] = font_dict;
    } else {
      WARN("PS stack overflow including MetaPost file or inline PS code");
      pdf_release_obj(font_dict);
      error = 1;
    }
  } else {
    error = 1;
  }

  return error;
}

static int
do_setfont (void)
{
  int      error = 0;
  int      font_id;
  char    *font_name;
  double   font_scale;
  pdf_obj *font_dict;

  font_dict = POP_STACK();
  if (!is_fontdict(font_dict))
    error = 1;
  else {
    font_name  = pdf_name_value(pdf_lookup_dict(font_dict, "FontName"));
    font_scale = pdf_number_value(pdf_lookup_dict(font_dict, "FontScale"));
    font_id = mp_locate_font(font_name, font_scale);
    mp_gstates[num_saves].font = font_id;
  }
  pdf_release_obj(font_dict);

  return error;
}

static int
do_currentfont (void)
{
  int error = 0;
  int font_id;
  pdf_obj *font_dict;

  font_id = CURRENT_FONT();
  if (font_id < 0 || font_id >= num_mp_fonts)
    return 1;
  else {
    font_dict = pdf_new_dict();
    pdf_add_dict(font_dict,
		 pdf_new_name("Type"),
		 pdf_new_name("Font"));
    pdf_add_dict(font_dict,
		 pdf_new_name("FontName"),
		 pdf_new_name(mp_fonts[font_id].tex_name));
    pdf_add_dict(font_dict,
		 pdf_new_name("FontScale"),
		 pdf_new_number(mp_fonts[font_id].pt_size));
    if (top_stack < PS_STACK_SIZE) {
      stack[top_stack++] = font_dict;
    } else {
      WARN("PS stack overflow including MetaPost file or inline PS code");
      pdf_release_obj(font_dict);
      error = 1;
    }
  }

  return error;
}

static int
do_show (void)
{
  pdf_xgstate *gs;
  double   font_scale;
  pdf_obj *text_str;
  int      length;
  unsigned char *strptr;
  double   text_width;

  gs = CURRENT_GS();

  text_str = POP_STACK();
  if (!PDF_OBJ_STRINGTYPE(text_str)) {
    if (text_str)
      pdf_release_obj(text_str);
    return 1;
  }
  if (gs->font < 0 || gs->font >= num_mp_fonts) {
    pdf_release_obj(text_str);
    return 1;
  }
  font_scale = mp_fonts[gs->font].pt_size;

  strptr = pdf_string_value (text_str);
  length = pdf_string_length(text_str);

  if (mp_fonts[gs->font].tfm_id < 0)
    WARN("MPOST: TFM not found for \"%s\".",
	 mp_fonts[gs->font].tex_name);

  text_width = 0.0;

  if (mp_fonts[gs->font].subfont_id >= 0) {
    unsigned short  uch;
    unsigned char  *ustr;
    int i;

    ustr = NEW(length*2, unsigned char);
    for (i = 0; i < length; i++) {
      uch = lookup_sfd_record(mp_fonts[gs->font].subfont_id, strptr[i]);
      ustr[2*i  ] = uch >> 8;
      ustr[2*i+1] = uch & 0xff;
      if (mp_fonts[gs->font].tfm_id >= 0) {
	text_width +=
	  tfm_get_width(mp_fonts[gs->font].tfm_id, strptr[i]);
      }
    }
    text_width *= font_scale;

    dev_set_string((spt_t)(gs->cp.x * dev_unit_dviunit()),
		   (spt_t)(gs->cp.y * dev_unit_dviunit()),
		   ustr, length / 2,
		   (spt_t)(text_width*dev_unit_dviunit()),
		   mp_fonts[gs->font].font_id, 0);
    RELEASE(ustr);
  } else {
#define FWBASE ((double) (1<<20))
    if (mp_fonts[gs->font].tfm_id >= 0) {
      text_width = (double) tfm_string_width(mp_fonts[gs->font].tfm_id, strptr, length)/FWBASE;
      text_width *= font_scale;
    }
    dev_set_string((spt_t)(gs->cp.x * dev_unit_dviunit()),
		   (spt_t)(gs->cp.y * dev_unit_dviunit()),
		   strptr, length,
		   (spt_t)(text_width*dev_unit_dviunit()),
		   mp_fonts[gs->font].font_id, 0);
  }

  if (pdf_dev_currentfont_wmode() == 1) {
    gs->cp.y -= text_width;
  } else {
    gs->cp.x += text_width;
  }

  {  /* FIXME */
    double cp[2];

    cp[0] = gs->cp.x;
    cp[1] = gs->cp.y;
    pdf_draw_append_path(CURRENT_PATH(), CURRENT_POINT(), PDF_PATH_MOVETO, cp);
  }
  graphics_mode();
  pdf_release_obj(text_str);

  return 0;
}

/*
 * string tfmname scale fshow -
 *
 * exch findfont exch scalefont setfont show
 */
static int
do_mpost_fshow (int opcode, double x_user, double y_user)
{
  int error = 0;
  static const char *ps_code_fshow = "exch findfont exch scalefont setfont show";
  char *start, *end;

  start = (char *) ps_code_fshow;
  end   = start + strlen(start);

  error = mp_parse_body(&start, end, x_user, y_user);

  return error;
}

static int
do_texfig_operator (int opcode, double x_user, double y_user)
{
  static transform_info fig_p;
  static int in_tfig = 0;
  double values[6];
  int error = 0;

  switch (opcode) {
  case STEXFIG:
    error = pop_get_numbers(values, 6);
    if (!error) {
      double   dvi2pts;
      pdf_rect cropbox;

      transform_info_clear(&fig_p);
      dvi2pts = 1.0/dev_unit_dviunit();

      fig_p.width  =  values[0] * dvi2pts;
      fig_p.height =  values[1] * dvi2pts;
      fig_p.u_llx  =  values[2] * dvi2pts;
      fig_p.u_lly  = -values[3] * dvi2pts;
      fig_p.u_urx  =  values[4] * dvi2pts;
      fig_p.u_ury  = -values[5] * dvi2pts;
      fig_p.user_bbox = 1;

      cropbox.llx = fig_p.u_llx;
      cropbox.lly = fig_p.u_lly;
      cropbox.urx = fig_p.u_urx;
      cropbox.ury = fig_p.u_ury;

      pdf_doc_start_grabbing(NULL, fig_p.u_llx, fig_p.u_ury, &cropbox);
      
      in_tfig = 1;
    }
    break;
  case ETEXFIG:
    {
      int      xobj_id;
      pdf_obj *xform;
      xform_info  info;

      if (!in_tfig)
	ERROR("endTexFig without valid startTexFig!.");

      pdf_ximage_init_form_info(&info);
      pdf_doc_current_form_info(NULL, &(info.bbox));

      xform   = pdf_doc_end_grabbing();

      xobj_id = pdf_ximage_defineresource("__TexFig",
					  PDF_XOBJECT_TYPE_FORM, &info, xform);
      pdf_ximage_put_image(xobj_id, &fig_p, x_user, y_user);

      in_tfig = 0;
    }
    break;
  default:
    error = 1;
  }

  return error;
}

/*
 * buggy...
 */

/*
 * Again, the only piece that needs x_user and y_user is
 * that piece dealing with texfig.
 */
static int
do_operator (const char *token, double x_user, double y_user)
{
  int error  = 0;
  int opcode = 0;
  pdf_xgstate *gs;
  double   values[12];
  pdf_obj *tmp = NULL;
  pdf_tmatrix matrix;
#define PUSH(o) { \
  if (top_stack < PS_STACK_SIZE) { \
    stack[top_stack++] = (o); \
  } else { \
    WARN("PS stack overflow including MetaPost file or inline PS code"); \
    error=1; \
    break;\
  } \
}

  opcode = get_opcode(token);

  switch (opcode) {
    /* Arithmetic operators */
  case ADD:
    error = pop_get_numbers(values, 2);
    if (!error)
      PUSH(pdf_new_number(values[0] + values[1]));
    break;
  case MUL:
    error = pop_get_numbers(values, 2);
    if (!error)
      PUSH(pdf_new_number(values[0]*values[1]));
    break;
  case NEG:
    error = pop_get_numbers(values, 1);
    if (!error)
      PUSH(pdf_new_number(-values[0]));
    break;
  case SUB:
    error = pop_get_numbers(values, 2);
    if (!error)
      PUSH(pdf_new_number(values[0] - values[1]));
    break;
  case DIV:
    error = pop_get_numbers(values, 2);
    if (!error)
      PUSH(pdf_new_number(values[0]/values[1]));
    break;
  case TRUNCATE: /* Round toward zero. */
    error = pop_get_numbers(values, 1);
    if (!error)
      PUSH(pdf_new_number(((values[0] > 0) ? floor(values[0]) : ceil(values[0]))));
    break;

    /* Stack operation */
  case CLEAR: error = do_clear();  break;
  case POP:
    tmp = POP_STACK();
    if (tmp)
      pdf_release_obj(tmp);
    break;
  case EXCH:  error = do_exch();   break;

    /* Path construction */
  case MOVETO:
    error = pop_get_numbers(values, 2);
    if (!error)
      error = pdf_draw_append_path(CURRENT_PATH(), CURRENT_POINT(),
				   PDF_PATH_MOVETO, values);
    break;
  case RMOVETO:
    error = pop_get_numbers(values, 2);
    if (!error)
      error = pdf_draw_append_path(CURRENT_PATH(), CURRENT_POINT(),
				   PDF_PATH_RMOVETO, values);
    break;
  case LINETO:
    error = pop_get_numbers(values, 2);
    if (!error)
      error = pdf_draw_append_path(CURRENT_PATH(), CURRENT_POINT(),
				   PDF_PATH_LINETO, values);
    break;
  case RLINETO:
    error = pop_get_numbers(values, 2);
    if (!error)
      error = pdf_draw_append_path(CURRENT_PATH(), CURRENT_POINT(),
				   PDF_PATH_RLINETO, values);
    break;
  case CURVETO:
    error = pop_get_numbers(values, 6);
    if (!error)
      error = pdf_draw_append_path(CURRENT_PATH(), CURRENT_POINT(),
				   PDF_PATH_CURVETO, values);
    break;
  case RCURVETO:
    error = pop_get_numbers(values, 6);
    if (!error)
      error = pdf_draw_append_path(CURRENT_PATH(), CURRENT_POINT(),
				   PDF_PATH_RCURVETO, values);
    break;
  case CLOSEPATH:
    error = pdf_draw_append_path(CURRENT_PATH(), CURRENT_POINT(),
				 PDF_PATH_CLOSE, values);
    break;
  case ARC:
    error = pop_get_numbers(values, 5);
    if (!error)
      error = pdf_draw_append_arc(CURRENT_PATH(), CURRENT_POINT(),
				  values[0], values[1],
				  values[2], values[2],
				  values[3], values[4]);
    break;
  case ARCN:
    error = pop_get_numbers(values, 5);
    if (!error)
      error = pdf_draw_append_arcn(CURRENT_PATH(), CURRENT_POINT(),
				   values[0], values[1],
				   values[2], values[2],
				   values[3], values[4]);
    break;
    
  case NEWPATH:
    pdf_draw_clear_path(CURRENT_PATH());
    pdf_doc_add_to_page(" n", 2);
    break;
  case STROKE:
    pdf_draw_path(CURRENT_PATH(), "S");
    pdf_draw_clear_path(CURRENT_PATH());
    break;
  case FILL:
    pdf_draw_path(CURRENT_PATH(), "f");
    pdf_draw_clear_path(CURRENT_PATH());
    break;

  case CLIP:
    pdf_draw_path(CURRENT_PATH(), "W");
    /* clip does not destroy current path */
    break;
  case EOCLIP:
    pdf_draw_path(CURRENT_PATH(), "W*");
    /* clip does not destroy current path */
    break;

    /* Graphics state operators: */
  case GSAVE:
    pdf_draw_gsave();
    error = do_gsave(); /* FIXME */
    break;
  case GRESTORE:
    pdf_draw_grestore();
    error = do_grestore(); /* FIXME */
    break;

  case CONCAT:
    tmp = POP_STACK();
    error = cvr_array(tmp, values, 6); /* This does pdf_release_obj() */
    tmp = NULL;
    if (error)
      WARN("Missing array before \"concat\".");
    else {
      pdf_setmatrix(&matrix,
		    values[0], values[1],
		    values[2], values[3],
		    values[4], values[5]);
      pdf_draw_transform(&matrix);
      __concat(&matrix); /* FIXME */
    }
    break;
  case SCALE:
    error = pop_get_numbers(values, 2);
    if (!error) {
      switch (mp_cmode) {
#ifndef WITHOUT_ASCII_PTEX
      case MP_CMODE_PTEXVERT:
	pdf_setmatrix(&matrix,
		      values[1], 0.0,
		      0.0      , values[0],
		      0.0      , 0.0);
	break;
#endif /* !WITHOUT_ASCII_PTEX */
      default:
	pdf_setmatrix(&matrix,
		      values[0], 0.0,
		      0.0      , values[1],
		      0.0      , 0.0);
	break;
      }

      pdf_draw_transform(&matrix);
      __concat(&matrix); /* FIXME */
    }
    break;
    /* Positive angle means clock-wise direction in graphicx-dvips??? */
  case ROTATE:
    error = pop_get_numbers(values, 1);
    if (!error) {
      values[0] = values[0] * M_PI / 180;

      switch (mp_cmode) {
      case MP_CMODE_DVIPSK:
      case MP_CMODE_MPOST: /* Really? */
#ifndef WITHOUT_ASCII_PTEX
      case MP_CMODE_PTEXVERT:
#endif /* !WITHOUT_ASCII_PTEX */
	pdf_setmatrix(&matrix,
		      cos(values[0]), -sin(values[0]),
		      sin(values[0]),  cos(values[0]),
		      0.0,             0.0);
	break;
      default:
	pdf_setmatrix(&matrix,
		      cos(values[0]) , sin(values[0]),
		      -sin(values[0]), cos(values[0]),
		      0.0,             0.0);
	break;
      }

      pdf_draw_transform(&matrix);
      __concat(&matrix); /* FIXME */
    }
    break;
  case TRANSLATE:
    error = pop_get_numbers(values, 2);
    if (!error) {
      pdf_setmatrix(&matrix,
		    1.0,       0.0,
		    0.0,       1.0,
		    values[0], values[1]);
      pdf_draw_transform(&matrix);
      __concat(&matrix); /* FIXME */
    }
    break;

  case SETDASH:
    error = pop_get_numbers(values, 1);
    if (!error) {
      pdf_obj *pattern, *dash;
      int i;

      gs = CURRENT_GS();
      gs->line_dash.offset = values[0];

      pattern = POP_STACK();
      if (!PDF_OBJ_ARRAYTYPE(pattern)) {
	if (pattern)
	  pdf_release_obj(pattern);
	error = 1;
	break;
      }
      gs->line_dash.num_dash = pdf_array_length(pattern);
      if (gs->line_dash.num_dash > PDF_DASH_SIZE_MAX) {
	WARN("Too many dashes...");
	pdf_release_obj(pattern);
	error = 1;
	break;
      }
      for (i = 0;
	   i < gs->line_dash.num_dash && !error ; i++) {
	dash = pdf_get_array(pattern, i);
	if (!PDF_OBJ_NUMBERTYPE(dash))
	  error = 1;
	else {
	  gs->line_dash.pattern[i] = pdf_number_value(dash);
	}
      }
      pdf_release_obj(pattern);
      if (!error) {
	pdf_draw_setdash(gs->line_dash.num_dash,
			 gs->line_dash.pattern, gs->line_dash.offset);
      }
    }
    break;
  case SETLINECAP:
    error = pop_get_numbers(values, 1);
    if (!error) {
      gs = CURRENT_GS();
      if (gs->line_cap != values[0]) {
	gs->line_cap = values[0];
	pdf_draw_setlinecap(values[0]);
      }
    }
    break;
  case SETLINEJOIN:
    error = pop_get_numbers(values, 1);
    if (!error) {
      gs = CURRENT_GS();
      if (gs->line_join != values[0]) {
	gs->line_join = values[0];
	pdf_draw_setlinejoin(values[0]);
      }
    }
    break;
  case SETLINEWIDTH:
    error = pop_get_numbers(values, 1);
    if (!error) {
      gs = CURRENT_GS();
      if (gs->line_width != values[0]) {
	gs->line_width = values[0];
	pdf_draw_setlinewidth(values[0]);
      }
    }
    break;
  case SETMITERLIMIT:
    error = pop_get_numbers(values, 1);
    if (!error) {
      gs = CURRENT_GS();
      if (gs->miter_limit != values[0]) {
	gs->miter_limit = values[0];
	pdf_draw_setmiterlimit(values[0]);
      }
    }
    break;

  case SETCMYKCOLOR:
    error = pop_get_numbers(values, 4);
    if (!error) {
      gs = CURRENT_GS();
      gs->color.space_id = PDF_COLOR_DEV_CMYK;
      gs->color.values[0] = values[0];
      gs->color.values[1] = values[1];
      gs->color.values[2] = values[2];
      gs->color.values[3] = values[3];
      pdf_color_setcolor(&gs->color, &gs->color);
    }
    break;
  case SETGRAY:
    error = pop_get_numbers(values, 1);
    if (!error) {
      gs = CURRENT_GS();
      gs->color.space_id  = PDF_COLOR_DEV_GRAY;
      gs->color.values[0] = values[0];
      pdf_color_setcolor(&gs->color, &gs->color);
    }
    break;
  case SETRGBCOLOR:
    error = pop_get_numbers(values, 3);
    if (!error) {
      gs = CURRENT_GS();
      gs->color.space_id  = PDF_COLOR_DEV_RGB;
      gs->color.values[0] = values[0];
      gs->color.values[1] = values[1];
      gs->color.values[2] = values[2];
      pdf_color_setcolor(&gs->color, &gs->color);
    }
    break;

  case SHOWPAGE: /* Let's ignore this for now */
    break;

  case CURRENTPOINT:
    gs = CURRENT_GS();
    PUSH(pdf_new_number(gs->cp.x));
    PUSH(pdf_new_number(gs->cp.y));
    break;

  case DTRANSFORM:
    {
      pdf_coord p;

      pdf_copymatrix(&matrix, CURRENT_MATRIX());

      tmp = POP_STACK();
      if (PDF_OBJ_ARRAYTYPE(tmp)) {
	error = cvr_array(tmp, values, 6); /* This does pdf_release_obj() */
	tmp = NULL;
	if (error)
	  break;
	pdf_setmatrix(&matrix,
		      values[0], values[1],
		      values[2], values[3],
		      values[4], values[5]);
	tmp = POP_STACK();
      }
      
      if (!PDF_OBJ_NUMBERTYPE(tmp)) {
	error = 1;
	break;
      }
      p.y = pdf_number_value(tmp);
      pdf_release_obj(tmp);

      tmp = POP_STACK();
      if (!PDF_OBJ_NUMBERTYPE(tmp)) {
	error = 1;
	break;
      }
      p.x = pdf_number_value(tmp);
      pdf_release_obj(tmp);

      pdf_draw_dtransform_coord(&p, &matrix);
      PUSH(pdf_new_number(p.x));
      PUSH(pdf_new_number(p.y));
    }
    break;

  case IDTRANSFORM:
    {
      pdf_coord p;

      pdf_copymatrix(&matrix, CURRENT_MATRIX());

      tmp = POP_STACK();
      if (PDF_OBJ_ARRAYTYPE(tmp)) {
	error = cvr_array(tmp, values, 6); /* This does pdf_release_obj() */
	tmp = NULL;
	if (error)
	  break;
	pdf_setmatrix(&matrix,
		      values[0], values[1],
		      values[2], values[3],
		      values[4], values[5]);
	tmp = POP_STACK();
      }
      
      if (!PDF_OBJ_NUMBERTYPE(tmp)) {
	error = 1;
	break;
      }
      p.y = pdf_number_value(tmp);
      pdf_release_obj(tmp);

      tmp = POP_STACK();
      if (!PDF_OBJ_NUMBERTYPE(tmp)) {
	error = 1;
	break;
      }
      p.x = pdf_number_value(tmp);
      pdf_release_obj(tmp);

      pdf_draw_idtransform_coord(&p, &matrix);
      PUSH(pdf_new_number(p.x));
      PUSH(pdf_new_number(p.y));
    }
    break;

  case FINDFONT:
    error = do_findfont();
    break;
  case SCALEFONT:
    error = do_scalefont();
    break;
  case SETFONT:
    error = do_setfont();
    break;
  case CURRENTFONT:
    error = do_currentfont();
    break;

  case SHOW:
    error = do_show();
    break;

    /* Extensions */
  case FSHOW:
    do_mpost_fshow(opcode, x_user, y_user);
    /* Treat fshow as a path terminator of sorts */
    break;
  case STEXFIG:
  case ETEXFIG:
    error = do_texfig_operator(opcode, x_user, y_user);
    break;

  default:
    if (mp_is_fontname(token)) {
      PUSH(pdf_new_name(token));
    } else {
      WARN("Unknown token \"%s\"", token);
      error = 1;
    }
    break;
  }

  return error;
}

/*
 * the only sections that need to know x_user and y _user are those
 * dealing with texfig.
 */
static int
mp_parse_body (char **start, char *end, double x_user, double y_user)
{
  char    *token;
  pdf_obj *obj;
  int      error = 0;

  skip_white(start, end);
  while (*start < end && !error) {
    if (isdigit(**start) ||
	(*start < end - 1 &&
	 (**start == '+' || **start == '-' || **start == '.' ))) {
      double value;
      char  *next;

      value = strtod(*start, &next);
      if (next < end && !strchr("<([{/%", *next) && !isspace(*next)) {
	WARN("Unkown PostScript operator.");
	dump(*start, next);
	error = 1;
      } else {
	PUSH(pdf_new_number(value));
	*start = next;
      }
      /*
       * PDF parser can't handle PS operator inside arrays.
       * This shouldn't use parse_pdf_array().
       */
    } else if (**start == '[' &&
	       (obj = parse_pdf_array(start, end))) {
      PUSH(obj);
      /* This cannot handle ASCII85 string. */
    } else if (*start < end - 1 &&
	       (**start == '<' && *(*start+1) == '<') &&
	       (obj = parse_pdf_dict(start, end))) {
      PUSH(obj);
    } else if ((**start == '(' || **start == '<') &&
	       (obj = parse_pdf_string (start, end))) {
      PUSH(obj);
#if 0
    } else if (**start == '/') {
      WARN("Unable to handle names in raw PS code.");
      dump(*start, end);
      error = 1;
#endif
    } else if (**start == '/' &&
	       (obj = parse_pdf_name(start, end))) {
      PUSH(obj);
    } else {
      token = parse_ident(start, end);
      if (!token)
	error = 1;
      else {
	error = do_operator(token, x_user, y_user);
	RELEASE(token);
      }
    }
    skip_white(start, end);
  }
  if (error || *start < end) {
    WARN("Remainder of line unparsed.");
    dump(*start, end);
  }

  return error;
}

void
mp_eop_cleanup (void)
{
  pdf_xgstate *gs;

  release_fonts();

  do_clear();

  while (num_saves >= 0) {
    gs = &(mp_gstates[num_saves]);
    if (gs->path)
      pdf_draw_release_path(gs->path);
    gs->path = NULL;
    num_saves--;
  }
}

static void
clear_state (void)
{
  top_stack = 0;
  init_gstate(&mp_gstates[num_saves]);
}

int
do_raw_ps_special (char **start, char* end,
		   int cleanup, double x_user, double y_user)
{
  int error;

  if (dev_wmode() == 1)
    mp_cmode = MP_CMODE_PTEXVERT;
  else
    mp_cmode = MP_CMODE_DVIPSK;

  if (num_saves < 0) {
    num_saves++;
    clear_state();
  }
  /* Comment in dvipdfm:
   * Remember that x_user and y_user are off by 0.02 %
   */
  set_currentpoint(x_user, y_user);

  dev_start_mp_mode();
  error = mp_parse_body(start, end, x_user, y_user);
  dev_end_mp_mode();

  return !error;
}

/* mp inclusion is a bit of a hack.  The routine
 * starts a form at the lower left corner of
 * the page and then calls begin_form_xobj telling
 * it to record the image drawn there and bundle it
 * up in an xojbect.  This allows us to use the coordinates
 * in the MP file directly.  This appears to be the
 * easiest way to be able to use the dev_set_string()
 * command (with its scaled and extended fonts) without
 * getting all confused about the coordinate system.
 * After the xobject is created, the whole thing can
 * be scaled any way the user wants */
 
int
mps_include_page (pdf_ximage *ximage, transform_info *p,
		  FILE *image_file)
{
  int error = 1;
  pdf_obj *xobject;
  xform_info info;
  double ref_x, ref_y;
  char *buffer, *start, *end;
  long  size;

  pdf_ximage_init_form_info(&info);

  rewind(image_file);
  if ((size = file_size(image_file)) == 0) {
    WARN("Can't read any byte in the MPS file.");
    return -1;
  }

  buffer = NEW(size+1, char);
  fread(buffer, sizeof(char), size, image_file);
  buffer[size] = 0;
  start = buffer;
  end   = buffer + size;

  error = scan_bbox(&start, end, &(info.bbox));
  if (error) {
    WARN("Errors occured while scanning MetaPost file headers.");
    RELEASE(buffer);
    return -1;
  }

  dev_start_mp_mode();
  mp_cmode = MP_CMODE_MPOST;

  /* The matrix of xform is adjusted so that ref_x and ref_y met
   * the referenct point...
   */
  ref_x = p->user_bbox ? p->u_llx : 0.0;
  ref_y = p->user_bbox ? p->u_lly : 0.0;

  pdf_doc_start_grabbing(NULL, ref_x, ref_y, &(info.bbox));

  /*
   * gsave + clear state
   */
  num_saves++;
  clear_state();

  skip_prolog(&start, end);
  error = mp_parse_body(&start, end, 0.0, 0.0);
  if (error) {
    WARN("Errors occured while interpreting MetaPost file.");
    xobject = pdf_doc_end_grabbing();
    if (xobject)
      pdf_release_obj(xobject);
  } else {
    xobject = pdf_doc_end_grabbing();
    pdf_ximage_set_form(ximage, &info, xobject);
  }

  dev_end_mp_mode();

  RELEASE(buffer);

  /*
   * The reason why we don't return XObject itself is
   * PDF inclusion may not be made so.
   */
  return (error ? -1 : 0);
}

int
mps_do_page (FILE *image_file)
{
  int   error = 0;
  pdf_rect bbox;
  char *buffer, *start, *end;
  long  size;

  rewind(image_file);
  if ((size = file_size(image_file)) == 0) {
    WARN("Can't read any byte in the MPS file.");
    return -1;
  }

  buffer = NEW(size+1, char);
  fread(buffer, sizeof(char), size, image_file);
  buffer[size] = 0;
  start = buffer;
  end   = buffer + size;

  error = scan_bbox(&start, end, &bbox);
#if 0
  dev_set_page_size(bbox.urx - bbox.llx, bbox.ury - bbox.lly);
#endif
  if (error) {
    WARN("Errors occured while scanning MetaPost file headers.");
    RELEASE(buffer);
    return -1;
  }

  mp_cmode = MP_CMODE_MPOST;

  dev_bop();
  pdf_doc_set_mediabox(pdf_doc_current_page_no(), &bbox);

  dev_start_mp_mode();

  num_saves++;
  clear_state();

  skip_prolog(&start, end);
  error = mp_parse_body(&start, end, 0.0, 0.0);
  if (error) {
    WARN("Errors occured while interpreting MetaPost file.");
  }
  dev_end_mp_mode();
  dev_eop();

  RELEASE(buffer);

  /*
   * The reason why we don't return XObject itself is
   * PDF inclusion may not be made so.
   */
  return (error ? -1 : 0);
}

int
check_for_mp (FILE *image_file) 
{
  int try_count = 10;

  rewind (image_file);
  mfgets(work_buffer, WORK_BUFFER_SIZE, image_file);
  if (strncmp(work_buffer, "%!PS", 4))
    return 0;

  while (try_count > 0) {
    mfgets(work_buffer, WORK_BUFFER_SIZE, image_file);
    if (!strncmp(work_buffer, "%%Creator:", 10)) {
      if (strlen(work_buffer+10) >= 8 &&
	  strstr(work_buffer+10, "MetaPost"))
	break;
    }
    try_count--;
  }

  return ((try_count > 0) ? 1 : 0);
}
