/*  $Header: /home/cvsroot/dvipdfmx/src/pdfcolor.c,v 1.1 2004/08/01 12:58:25 hirata Exp $
    
    This is dvipdfmx, an eXtended version of dvipdfm by Mark A. Wicks.

    Copyright (C) 2002 by Jin-Hwan Cho and Shunsaku Hirata,
    the dvipdfmx project team <dvipdfmx@project.ktug.or.kr>
    
    Copyright (C) 1998, 1999 by Mark A. Wicks <mwicks@kettering.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "system.h"
#include "mfileio.h"
#include "mem.h"
#include "error.h"

#include "pdfdoc.h"
#include "pdfdev.h"

#include "colors.h"
#include "pdfcolor.h"

int
pdf_color_by_name (pdf_color *color, const char *color_name)
{
  int i;

  for (i = 0; i < NUM_NAMED_COLORS; i++) {
    if (!strcasecmp(colors_by_name[i].key, color_name)) {
      memcpy(color, &(colors_by_name[i].value), sizeof(pdf_color));
      return 0;
    }
  }
  if (i == NUM_NAMED_COLORS)
    WARN("Color \"%s\" not known.", color_name);

  return -1;
}


pdf_color bground_color = {
  PDF_COLOR_DEV_GRAY,
  {1.0, 1.0, 1.0, 1.0}
};

pdf_color default_color = {
  PDF_COLOR_DEV_GRAY,
  {0.0, 0.0, 0.0, 0.0}
};

pdf_color current_fill   = {
  PDF_COLOR_DEV_GRAY,
  {0.0, 0.0, 0.0, 0.0}
};

pdf_color current_stroke = {
  PDF_COLOR_DEV_GRAY,
  {0.0, 0.0, 0.0, 0.0}
};


/*
 * This routine is not a real color matching.
 */
static int
color_match (const pdf_color *color1, const pdf_color *color2)
{
  if (color1->space_id != color2->space_id)
    return 0;

  switch (color1->space_id) {
  case PDF_COLOR_DEV_RGB:
    if (color1->values[0] == color2->values[0] &&
	color1->values[1] == color2->values[1] &&
	color1->values[2] == color2->values[2])
      return 1;
    break;
  case PDF_COLOR_DEV_CMYK:
    if (color1->values[0] == color2->values[0] &&
	color1->values[1] == color2->values[1] &&
	color1->values[2] == color2->values[2] &&
	color1->values[3] == color2->values[3])
      return 1;
    break;
  case PDF_COLOR_DEV_GRAY:
    if (color1->values[0] == color2->values[0])
      return 1;
    break;
  }

  return 0;
}

pdf_color *
pdf_color_default (void)
{
  return &default_color;
}

pdf_color *
pdf_color_bgcolor (void)
{
  return &bground_color;
}

pdf_color *
pdf_color_current_stroke (void)
{
  return &current_stroke;
}

pdf_color *
pdf_color_current_fill (void)
{
  return &current_fill;
}

int
pdf_sprint_color_values (char *format_buffer, const pdf_color *color)
{
  int len = 0;

  switch (color->space_id) {
  case PDF_COLOR_DEV_RGB:
    len = sprintf(format_buffer, " %.2f %.2f %.2f",
		  color->values[0], color->values[1], color->values[2]);
    break;
  case PDF_COLOR_DEV_CMYK:
    len = sprintf(format_buffer, " %.2f %.2f %.2f %.2f",
		  color->values[0], color->values[1],
		  color->values[2], color->values[3]);
    break;
  case PDF_COLOR_DEV_GRAY:
    len = sprintf(format_buffer, " %.2f", color->values[0]);
    break;
  case PDF_COLOR_INVALID:
    ERROR("Invalid color item.");
    break;
  default:
    ERROR("Only RGB/CMYK/Gray currently supported.");
  }

  return len;
}

void
pdf_color_setcolor (const pdf_color *color_stroke, const pdf_color *color_fill)
{
  int len = 0;
  char format_buffer[1024];

  if (color_stroke) {
    len = pdf_sprint_color_values(format_buffer, color_stroke);
    pdf_doc_add_to_page(format_buffer, len);
    switch (color_stroke->space_id) {
    case PDF_COLOR_DEV_RGB:
      pdf_doc_add_to_page(" RG", 3);
      break;
    case PDF_COLOR_DEV_CMYK:
      pdf_doc_add_to_page(" K", 2);
      break;
    case PDF_COLOR_DEV_GRAY:
      pdf_doc_add_to_page(" G", 2);
      break;
    case PDF_COLOR_INVALID:
      ERROR("Invalid color item.");
      break;
    default:
      ERROR("Only RGB/CMYK/Gray currently supported.");
      pdf_doc_add_to_page(" SC", 3);
    }
  }

  if (color_fill) {
    len = pdf_sprint_color_values(format_buffer, color_fill);
    pdf_doc_add_to_page(format_buffer, len);
    switch (color_fill->space_id) {
    case PDF_COLOR_DEV_RGB:
      pdf_doc_add_to_page(" rg", 3);
      break;
    case PDF_COLOR_DEV_CMYK:
      pdf_doc_add_to_page(" k", 2);
      break;
    case PDF_COLOR_DEV_GRAY:
      pdf_doc_add_to_page(" g", 2);
      break;
    case PDF_COLOR_INVALID:
      ERROR("Invalid color item.");
      break;
    default:
      ERROR("Only RGB/CMYK/Gray currently supported.");
      pdf_doc_add_to_page(" sc", 3);
    }
  }

  if (color_stroke)
    memcpy(&current_stroke, color_stroke, sizeof(pdf_color));
  if (color_fill)
    memcpy(&current_fill,   color_fill,   sizeof(pdf_color));
}

void
pdf_color_set_bgcolor (const pdf_color *color)
{
  bground_color.space_id  = color->space_id;
  bground_color.values[0] = color->values[0];
  bground_color.values[1] = color->values[1];
  bground_color.values[2] = color->values[2];
  bground_color.values[3] = color->values[3];
}

void
pdf_color_set_default (const pdf_color *color)
{
  memcpy(&default_color, color, sizeof(pdf_color));
}

/*
 * Having color stack here is wrong.
 * dvi.c or XXX_special should treat this.
 */
#define DEV_COLOR_STACK_MAX 256

static struct {
  int top;
  pdf_color colors[DEV_COLOR_STACK_MAX];
} color_stack;

#define CURRENTCOLOR() (color_stack.top > 0 ? &(color_stack.colors[color_stack.top-1]) : &default_color)
#define PREVCOLOR()    (color_stack.top > 1 ? &(color_stack.colors[color_stack.top-2]) : &default_color)

void
pdf_color_stack_clear (void)
{
  color_stack.top = 0;
}

pdf_color *
pdf_color_stack_top (void)
{
  return CURRENTCOLOR();
}

void
pdf_color_stack_push (const pdf_color *color)
{
  pdf_color *currentcolor, *prevcolor;

  if (color_stack.top >= DEV_COLOR_STACK_MAX) {
    WARN("Exceeded depth of color stack.");
    return;
  }

  prevcolor    = CURRENTCOLOR();
  currentcolor = &(color_stack.colors[color_stack.top]);

  memcpy(currentcolor, color, sizeof(pdf_color));
  color_stack.top += 1;

  /*
   * This may not reselect current color when the color is changed in
   * the pdf:literal special. But users are completely responsible for
   * doing correct thing in that case; any pdf:literal must be enclosed
   * by "q" and "Q" manually.
   */
  if (!color_match(currentcolor, prevcolor)) {
    pdf_color_setcolor(currentcolor, currentcolor);
  }
}

void
pdf_color_stack_pop (void)
{
  pdf_color *prevcolor;

  if (color_stack.top <= 0) {
    WARN("End color with no corresponding begin color.");
  } else {
    prevcolor = PREVCOLOR();
    if (!color_match(CURRENTCOLOR(), prevcolor)) {
      pdf_color_setcolor(prevcolor, prevcolor);
    }
    color_stack.top -= 1;
  }
}

/***************************** COLOR SPACE *****************************/
